#!/usr/bin/env python3
"""
Load the model, then do inference on the test data
"""

import time
import os
import sys

import numpy as np
import torch
import torchvision
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import datasets, transforms
from tncom.args import args
from tncom import conv
from tncom import mpofc
from tncom import mpomlp
from tncom import opmlp
from tncom.utils import mylog
import tncom.pepo as pepo
from tncom.mpolinear import MPOLinear3

def test(args, model, test_loader):
    model.eval()
    test_loss = 0
    correct = 0
    n_test = 0
    with torch.no_grad():
        for data, target in test_loader:
            if not args.ingpu:
                data, target = data.to(args.device), target.to(args.device)
            output = model(data)
            test_loss += F.nll_loss(output, target, reduction='sum').item() # sum up batch loss
            pred = output.argmax(dim=1, keepdim=True) # get the index of the max log-probability
            correct += pred.eq(target.view_as(pred)).sum().item()
            n_test += args.test_batch_size

    test_loss /= n_test
    correct /= n_test
    return test_loss,correct

def main():
    start_time = time.time()
    assert(args.load)
    t0 = time.time()
    sys.stdout.write("Loading form %s..."%args.load);sys.stdout.flush()
    [myargs,net,test_data] = torch.load(args.load)
    print("done, %.2f Sec."%(time.time() - t0))
    print("-"*80)
    print("Model architecture:")
    print(net.parameters)

    print("-"*80)
    print("Size of parameters:")
    params = list(net.parameters())

    for name, param in net.named_parameters():
        print(name," shape =",list(param.size()),"Containing",np.prod(param.shape),"parameters")
    params = list(filter(lambda p: p.requires_grad, params))
    nparams = int(sum([np.prod(p.shape) for p in params]))
    mylog('Total number of trainable parameters: {}'.format(nparams))
    named_params = list(net.named_parameters())

    print("-"*80)
    print("Start inference...")
    if(args.cpu):
        net = net.to('cpu')
        print("Infernce using CPU")
        for batch in range(len(test_data)):
            test_data[batch] = (test_data[batch][0].to('cpu'),test_data[batch][1].to('cpu'))
    else:
        print("Infernce using GPU",args.cuda)
        if args.device !=  myargs.device:
            net = net.to(args.device)
            for batch in range(len(test_data)):
                test_data[batch] = (test_data[batch][0].to(args.device),test_data[batch][1].to(args.device))
    t0_test = time.time()
    loss_test,accu_test=test(myargs, net, test_data)
    t_test = time.time()-t0_test
    print("test_accuracy = %.4f   Time used: %.2f Sec."%(accu_test,t_test))

    print("Rounding, cutoff=",args.cutoff)
    for idx, m in enumerate(net.modules()):
        if isinstance(m,MPOLinear3) and idx == 1:
            m.dmrg(args.cutoff)
            #m.cano()
    t0_test = time.time()
    loss_test,accu_test=test(myargs, net, test_data)
    t_test = time.time()-t0_test
    print("test_accuracy = %.4f   Time used: %.2f Sec."%(accu_test,t_test))
    print("-"*80)


if __name__ == '__main__':
    main()
