import torch.nn as nn
import torch.nn.functional as F
import torch
from .args import args
import numpy as np
from . import mpomlp
from . import opmlp

class Net(nn.Module):
    def __init__(self,args):
        super(Net, self).__init__()
        self.channels = [int(i) for i in args.channels.split(",")]
        self.args = args
        self.paddings = [int(i) for i in args.paddings.split(",")]
        assert len(self.channels) == len(self.paddings)
        self.n_channel_1=16
        self.n_channel_2=64
        padding_1=0
        padding_2=2
        self.fc_size=6
        self.hidden = 256
        self.conv1 = nn.Conv2d(1, self.n_channel_1, 5, 1,padding=padding_1)
        self.conv2 = nn.Conv2d(self.n_channel_1, self.n_channel_2, 5, 1,padding=padding_2)
        self.fc1 = nn.Linear(self.fc_size*self.fc_size*self.n_channel_2, self.hidden)
        self.fc2 = nn.Linear(self.hidden, 10)
        if args.bn:
            self.bn1 = nn.BatchNorm2d(num_features=self.n_channel_1)
            self.bn2 = nn.BatchNorm2d(num_features=self.n_channel_2)
        if args.dropout:
            self.dropout1 = nn.Dropout(0.2)
            self.dropout2 = nn.Dropout(0.2)

        params = list(self.conv1.parameters())
        params = list(filter(lambda p: p.requires_grad, params))
        nparams = int(sum([np.prod(p.shape) for p in params]))
        print('# parameters in Conv1: {}'.format(nparams))
        params = list(self.conv2.parameters())
        params = list(filter(lambda p: p.requires_grad, params))
        nparams = int(sum([np.prod(p.shape) for p in params]))
        print('# parameters in Conv2: {}'.format(nparams))

    def forward(self, x):
        x = F.relu(self.conv1(x))
        if self.args.bn:
            x = self.bn1(x)
        x = F.max_pool2d(x, 2, 2)
        if self.args.dropout:
            x = self.dropout1(x)
        x = F.relu(self.conv2(x))
        if self.args.bn:
            x = self.bn2(x)
        x = F.max_pool2d(x, 2, 2) ## after this step, the data has shape [ batch, channel, 6, 6]
        if self.args.dropout:
            x = self.dropout2(x)
        x = x.view(-1, self.fc_size*self.fc_size*self.n_channel_2)
        x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)

class Net2(nn.Module):
    def __init__(self,args):
        super(Net2, self).__init__()
        self.args = args
        self.channels = [int(i) for i in args.channels.split(",")]
        self.paddings = [int(i) for i in args.paddings.split(",")]
        self.kernel_size = [int(i) for i in args.kernel_size.split(",")]
        print("channels:",self.channels)
        print("paddings:",self.paddings)
        print("width:",self.kernel_size)
        assert len(self.channels) == len(self.paddings) 
        assert len(self.channels) == len(self.kernel_size) 
        self.net = nn.Sequential()
        self.size = self.args.size
        if args.seed > 0:
            torch.manual_seed(args.seed)
        for i in range(len(self.channels)):
            ch0 = 3 if args.dataset == 'cifar10' else 1
            ch1 = ch0 if i==0 else self.channels[i-1]
            ch2 = self.channels[i]
            padding = self.paddings[i]
            kernel_size = self.kernel_size[i]
            self.net.add_module("conv%d"%i,nn.Conv2d(ch1,ch2,kernel_size,1,padding=padding))
            self.net.add_module("relu%d"%i,nn.ReLU())
            if self.args.bn:
                nn.BatchNorm2d(num_features=ch2)
            self.net.add_module("pull%d"%i,nn.MaxPool2d(2))
            if self.args.dropout > 0:
                self.net.add_module("dropout%d"%i,nn.Dropout(self.args.dropout))
            self.size = (self.size - kernel_size + 1 +padding*2)//2
        print("final conv output size=",self.size)

        self.net2 = nn.Sequential()
        self.fc_size = self.size*self.size*self.channels[-1]
        if self.args.convmpo:
            self.net2.add_module("mpo",mpomlp.Net(self.args,indim = self.fc_size))
        elif self.args.convop:
            self.net2.add_module("op",opmlp.Net(self.args,indim = self.fc_size))
        else:
            self.hidden = self.args.D
            self.net2.add_module("linear1",nn.Linear(self.fc_size, self.hidden))
            self.net2.add_module("relu_linear",nn.ReLU())
            self.net2.add_module("linear2",nn.Linear(self.hidden, 10))

        params = list(self.net.parameters())
        params = list(filter(lambda p: p.requires_grad, params))
        nparams = int(sum([np.prod(p.shape) for p in params]))
        print('# parameters in Conv_net: {}'.format(nparams))

        params = list(self.net2.parameters())
        params = list(filter(lambda p: p.requires_grad, params))
        nparams = int(sum([np.prod(p.shape) for p in params]))
        print('# parameters in MLP: {}'.format(nparams))

    def forward(self, x):
        if self.args.convmpo or self.args.convop:
            return self.net2(self.net(x))
        else:
            batch_size = x.shape[0]
            x = self.net(x)
            x = self.net2(x.view(batch_size,-1))
            return F.log_softmax(x, dim=1)



